@@ -14,10 +14,13 @@
 //!     of leases that resolve anytime there is an available [`Lease`]
 
 #![cfg_attr(not(test), deny(warnings, clippy::all, clippy::pedantic, clippy::cargo))]
+#![allow(clippy::single_match_else)]
 #![deny(unsafe_code, missing_docs)]
 use self::wrapper::Wrapper;
+use core::future::Future;
+use core::iter::FromIterator;
 use core::ops::{Deref, DerefMut};
-use std::iter::FromIterator;
+use parking_lot::Mutex;
 use std::sync::Arc;
 
 #[cfg(feature = "async")]
@@ -39,14 +42,29 @@ pub struct Pool<T> {
 }
 
 struct PoolInner<T> {
-  buffer: lockfree::set::Set<Arc<Wrapper<T>>>,
+  buffer: lockfree::set::Set<Wrapper<T>>,
   #[cfg(feature = "async")]
   waiting_futures: async_::WaitingFutures,
 }
 
+impl<T> Default for PoolInner<T> {
+  fn default() -> Self {
+    Self {
+      buffer: lockfree::set::Set::default(),
+      #[cfg(feature = "async")]
+      waiting_futures: async_::WaitingFutures::default(),
+    }
+  }
+}
+
 impl<T> Pool<T> {
+  /// Creates a new empty `Pool`
+  pub fn new() -> Self {
+    Self::default()
+  }
+
   /// Creates a new `Pool` with an initial size of `pool_size` by calling `init` `pool_size` times.
-  pub fn new(pool_size: usize, mut init: impl FnMut() -> T) -> Self {
+  pub fn with_initial_size(pool_size: usize, mut init: impl FnMut() -> T) -> Self {
     (0..pool_size).map(|_| init()).collect()
   }
 
@@ -54,10 +72,12 @@ impl<T> Pool<T> {
   ///
   /// # Errors
   /// This returns the very first error returned by `init`
-  pub fn try_new<E>(pool_size: usize, mut init: impl FnMut() -> Result<T, E>) -> Result<Self, E> {
+  pub fn try_with_initial_size<E>(pool_size: usize, mut init: impl FnMut() -> Result<T, E>) -> Result<Self, E> {
     let buffer = lockfree::set::Set::new();
     for _ in 0..pool_size {
-      buffer.insert(Arc::new(Wrapper::new(init()?))).ok();
+      buffer
+        .insert(Wrapper::new(init()?))
+        .unwrap_or_else(|_| unreachable!("Each new wrapper should be unique"));
     }
     Ok(Self {
       inner: Arc::new(PoolInner {
@@ -68,92 +88,141 @@ impl<T> Pool<T> {
     })
   }
 
+  /// Tries to get a [`Lease`] if one is available. This function does not block.
+  ///
+  /// For an asynchronous version that returns when one is available use [`get_async()`](Self::get_async())
+  #[must_use]
+  pub fn get(&self) -> Option<Lease<T>> {
+    self.inner.buffer.iter().find_map(|wrapper| Lease::from_arc_mutex(&wrapper, self))
+  }
+
   /// Returns a future that resolves to a [`Lease`] when one is available
+  ///
+  /// Reqires the `async` feature to be enabled because it requires extra memory
   #[cfg(feature = "async")]
   pub fn get_async(&self) -> async_::AsyncLease<T> {
     async_::AsyncLease::new(self)
   }
 
-  /// Tries to get a [`Lease`] if one is available. This function does not block.
-  #[must_use]
-  pub fn get(&self) -> Option<Lease<T>> {
-    self.inner.buffer.iter().find_map(|arc| Lease::from_arc_mutex(&arc, self))
-  }
-
   /// Returns a struct that implements the [`futures_core::Stream`] trait.
   #[cfg(feature = "async")]
   pub fn stream(&self) -> async_::PoolStream<T> {
     async_::PoolStream::new(self)
   }
 
+  /// For the asynchronous version of this function see [`get_or_new_async()`](Self::get_or_new_async())
+  ///
   /// Tries to get an existing [`Lease`] if available and if not returns a new one that has been added to the pool.
   ///
   /// Calling this method repeatedly can cause the pool size to increase without bound.
-  #[allow(clippy::missing_panics_doc)]
   pub fn get_or_new(&self, init: impl FnOnce() -> T) -> Lease<T> {
-    self.get().map_or_else(
-      || {
-        let mutex = Arc::new(Wrapper::new(init()));
-        let lease = Lease::from_arc_mutex(&mutex, self).unwrap();
-        self.associate(&lease);
-        lease
-      },
-      |t| t,
-    )
+    self.get().map_or_else(|| self.insert_with_lease(init()), |t| t)
   }
 
+  /// Asynchronous version of [`get_or_new()`](Self::get_or_new())
+  ///
+  /// Tries to get an existing [`Lease`] if available and if not returns a new one that has been added to the pool.
+  ///
+  /// Calling this method repeatedly can cause the pool size to increase without bound.
+  pub async fn get_or_new_async(&self, init: impl Future<Output = T>) -> Lease<T> {
+    match self.get() {
+      Some(lease) => lease,
+      None => self.insert_with_lease(init.await),
+    }
+  }
+
+  /// For the asynchronous version of this function see [`get_or_try_new_async()`](Self::get_or_try_new_async())
+  ///
   /// Tries to get an existing [`Lease`] if available and if not tries to create a new one that has been added to the pool.
   ///
   /// Calling this method repeatedly can cause the pool size to increase without bound.
   ///
   /// # Errors
   /// Returns an error if `init` errors
-  #[allow(clippy::missing_panics_doc)]
   pub fn get_or_try_new<E>(&self, init: impl FnOnce() -> Result<T, E>) -> Result<Lease<T>, E> {
     match self.get() {
-      None => {
-        let mutex = Arc::new(Wrapper::new(init()?));
-        let lease = Lease::from_arc_mutex(&mutex, self).unwrap();
-        self.associate(&lease);
-        Ok(lease)
-      }
+      None => Ok(self.insert_with_lease(init()?)),
       Some(l) => Ok(l),
     }
   }
 
-  /// Just like [`get_or_new()`](Self::get_or_new()) but caps the size of the pool. Once [`len()`](Self::len()) == `cap` then `None` is returned.
-  #[allow(clippy::missing_panics_doc)]
-  pub fn get_or_new_with_cap(&self, cap: usize, init: impl FnOnce() -> T) -> Option<Lease<T>> {
-    if let Some(t) = self.get() {
-      Some(t)
-    } else {
-      if self.len() >= cap {
-        return None;
-      }
-      let mutex = Arc::new(Wrapper::new(init()));
-      let lease = Lease::from_arc_mutex(&mutex, self).unwrap();
-      self.associate(&lease);
-      Some(lease)
+  /// Asynchronous version of [`get_or_try_new()`](Self::get_or_try_new())
+  ///
+  /// Tries to get an existing [`Lease`] if available and if not tries to create a new one that has been added to the pool.
+  ///
+  /// Calling this method repeatedly can cause the pool size to increase without bound.
+  ///
+  /// # Errors
+  /// Returns an error if `init` errors
+  pub async fn get_or_try_new_async<E>(&self, init: impl Future<Output = Result<T, E>>) -> Result<Lease<T>, E> {
+    match self.get() {
+      None => Ok(self.insert_with_lease(init.await?)),
+      Some(l) => Ok(l),
     }
   }
 
+  /// For the asynchronous version of this function see [`get_or_new_with_cap_async()`](Self::get_or_new_with_cap_async())
+  ///
+  /// Just like [`get_or_new()`](Self::get_or_new()) but caps the size of the pool. Once [`len()`](Self::len()) == `cap` then `None` is returned.
+  pub fn get_or_new_with_cap(&self, cap: usize, init: impl FnOnce() -> T) -> Option<Lease<T>> {
+    match self.get() {
+      Some(t) => Some(t),
+      None => (self.len() < cap).then(|| self.insert_with_lease(init())),
+    }
+  }
+
+  /// Asynchronous version of [`get_or_new_with_cap()`](Self::get_or_new_with_cap())
+  ///
+  /// Just like [`get_or_new()`](Self::get_or_new()) but caps the size of the pool. Once [`len()`](Self::len()) == `cap` then `None` is returned.
+  pub async fn get_or_new_with_cap_async(&self, cap: usize, init: impl Future<Output = T>) -> Option<Lease<T>> {
+    match self.get() {
+      Some(t) => Some(t),
+      None => {
+        if self.len() < cap {
+          return None;
+        }
+        Some(self.insert_with_lease(init.await))
+      }
+    }
+  }
+
+  /// For the asynchronous version of this function see [`get_or_try_new_with_cap_async()`](Self::get_or_try_new_with_cap_async())
+  ///
   /// Just like [`get_or_try_new()`](Self::get_or_try_new()) but caps the size of the pool. Once [`len()`](Self::len()) == `cap` then `None` is returned.
   ///
   /// # Errors
   /// Returns an error if `init` errors
-  #[allow(clippy::missing_panics_doc)]
   pub fn get_or_try_new_with_cap<E>(&self, cap: usize, init: impl FnOnce() -> Result<T, E>) -> Result<Option<Lease<T>>, E> {
-    if let Some(t) = self.get() {
-      Ok(Some(t))
-    } else {
-      if self.len() >= cap {
-        return Ok(None);
+    match self.get() {
+      Some(t) => Ok(Some(t)),
+      None => {
+        if self.len() >= cap {
+          return Ok(None);
+        }
+        Ok(Some(self.insert_with_lease(init()?)))
       }
+    }
+  }
 
-      let mutex = Arc::new(Wrapper::new(init()?));
-      let lease = Lease::from_arc_mutex(&mutex, self).unwrap();
-      self.associate(&lease);
-      Ok(Some(lease))
+  /// Asynchronous version of [`get_or_try_new_with_cap()`](Self::get_or_try_new_with_cap())
+  ///
+  /// Just like [`get_or_try_new()`](Self::get_or_try_new()) but caps the size of the pool. Once [`len()`](Self::len()) == `cap` then `None` is returned.
+  ///
+  /// # Errors
+  /// Returns an error if `init` errors
+  pub async fn get_or_try_new_with_cap_async<E>(
+    &self,
+    cap: usize,
+    init: impl Future<Output = Result<T, E>>,
+  ) -> Result<Option<Lease<T>>, E> {
+    match self.get() {
+      Some(t) => Ok(Some(t)),
+      None => {
+        if self.len() >= cap {
+          return Ok(None);
+        }
+        Ok(Some(self.insert_with_lease(init.await?)))
+      }
     }
   }
 
@@ -169,9 +238,9 @@ impl<T> Pool<T> {
   /// holding will be dropped
   pub fn clear(&self) {
     self.inner.buffer.iter().for_each(|g| {
-      let arc: &Arc<_> = &*g;
-      self.inner.buffer.remove(arc);
-    })
+      let wrapper: &Wrapper<_> = &g;
+      self.inner.buffer.remove(wrapper);
+    });
   }
 
   /// Resizes the pool to `pool_size`
@@ -182,7 +251,7 @@ impl<T> Pool<T> {
     self.inner.buffer.iter().skip(pool_size).for_each(|g| {
       self.inner.buffer.remove(&*g);
     });
-    set.extend((self.len()..pool_size).map(|_| Arc::new(Wrapper::new(init()))));
+    set.extend((self.len()..pool_size).map(|_| Wrapper::new(init())));
   }
 
   /// Resizes the pool to `pool_size`
@@ -197,18 +266,45 @@ impl<T> Pool<T> {
       set.remove(&*g);
     });
     for _ in self.len()..pool_size {
-      set.insert(Arc::new(Wrapper::new(init()?))).ok();
+      set
+        .insert(Wrapper::new(init()?))
+        .unwrap_or_else(|_| unreachable!("Each new wrapper should be unique"));
     }
     Ok(())
   }
 
-  /// Adds the [`Lease`] to this [`Pool`] if it isn't already part of the pool.
-  pub fn associate(&self, lease: &Lease<T>) {
-    self.inner.buffer.insert(lease.mutex.clone()).ok();
+  /// Just like the [`Extend`](core::iter::Extend) trait but doesn't require self to be mutable
+  pub fn extend<I: IntoIterator<Item = T>>(&self, iter: I) {
+    self.inner.buffer.extend(iter.into_iter().map(Wrapper::new))
   }
 
-  /// Returns the number of currently available [`Lease`]es. Even if the return is non-zero calling [`get()`](Self::get())
-  /// immediately afterward can still fail if multiple.
+  /// Adds new item to this [`Pool`]
+  /// Use [`insert_with_lease()`](Self::insert_with_lease()) if you need a [`Lease`] back
+  pub fn insert(&self, t: T) {
+    let wrapper = Wrapper::new(t);
+    self
+      .inner
+      .buffer
+      .insert(wrapper)
+      .unwrap_or_else(|_| unreachable!("Each new wrapper should be unique"));
+    self.notify();
+  }
+
+  /// Adds new item to this [`Pool`] and returns a [`Lease`] that is ready to be used.
+  /// Use [`insert()`](Self::insert()) if you don't need a [`Lease`] back
+  pub fn insert_with_lease(&self, t: T) -> Lease<T> {
+    let wrapper = Wrapper::new(t);
+    let lease = Lease::from_arc_mutex(&wrapper, self).unwrap_or_else(|| unreachable!("Wrapper is unlocked when new"));
+    self
+      .inner
+      .buffer
+      .insert(wrapper)
+      .unwrap_or_else(|_| unreachable!("Each new wrapper should be unique"));
+    lease
+  }
+
+  /// Returns the number of currently available [`Lease`]es. Even if the return is non-zero, calling [`get()`](Self::get())
+  /// immediately afterward can still fail if multiple threads have access to this pool.
   #[must_use]
   pub fn available(&self) -> usize {
     self.inner.buffer.iter().filter(|b| !b.is_locked()).count()
@@ -220,9 +316,16 @@ impl<T> Pool<T> {
     self.inner.buffer.iter().next().is_none()
   }
 
-  /// Disassociates the returned ['Lease'] from this [`Pool`]
+  /// Disassociates the returned [`Lease`] from this [`Pool`]
   pub fn disassociate(&self, lease: &Lease<T>) {
-    self.inner.buffer.remove(&lease.mutex);
+    // This is the one unfortunate place where wrapping the arc is more costly.
+    // Since this function shouldn't be called in a tight loop, the clone should be fine
+    self.inner.buffer.remove(&Wrapper(lease.mutex.clone()));
+  }
+
+  fn notify(&self) {
+    #[cfg(feature = "async")]
+    self.inner.waiting_futures.wake_next();
   }
 }
 
@@ -240,20 +343,22 @@ impl<T: Default> Pool<T> {
 
   /// Just like [`resize()`](Self::resize()) but uses [`Default::default()`] as the `init` function
   pub fn resize_default(&self, pool_size: usize) {
-    self.resize(pool_size, T::default)
+    self.resize(pool_size, T::default);
   }
 }
 
 impl<T> Default for Pool<T> {
   fn default() -> Self {
-    Self::new(0, || unreachable!())
+    Self {
+      inner: Arc::new(PoolInner::default()),
+    }
   }
 }
 
 impl<T> core::fmt::Debug for Pool<T> {
   fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
     struct ListDebugger<'a, T> {
-      set: &'a lockfree::set::Set<Arc<Wrapper<T>>>,
+      set: &'a lockfree::set::Set<Wrapper<T>>,
     }
 
     impl<T> core::fmt::Debug for ListDebugger<'_, T> {
@@ -263,9 +368,10 @@ impl<T> core::fmt::Debug for Pool<T> {
     }
 
     let mut s = f.debug_struct("Pool");
-    s.field("len", &self.len()).field("available", &self.available());
-    s.field("availabilities", &ListDebugger { set: &self.inner.buffer });
-    s.finish()
+    s.field("len", &self.len())
+      .field("available", &self.available())
+      .field("availabilities", &ListDebugger { set: &self.inner.buffer })
+      .finish()
   }
 }
 
@@ -279,7 +385,7 @@ impl<T> FromIterator<T> for Pool<T> {
   fn from_iter<I: IntoIterator<Item = T>>(iter: I) -> Self {
     Self {
       inner: Arc::new(PoolInner {
-        buffer: iter.into_iter().map(Wrapper::new).map(Arc::new).collect(),
+        buffer: iter.into_iter().map(Wrapper::new).collect(),
         #[cfg(feature = "async")]
         waiting_futures: async_::WaitingFutures::new(),
       }),
@@ -304,7 +410,7 @@ fn assert_lease_is_static() {
 /// does so those can also be used to get access to the underlying data.  
 #[must_use]
 pub struct Lease<T> {
-  mutex: Arc<Wrapper<T>>,
+  mutex: Arc<Mutex<T>>,
   #[cfg(feature = "async")]
   pool: Pool<T>,
 }
@@ -318,14 +424,12 @@ impl<T> Drop for Lease<T> {
       self.mutex.force_unlock();
     }
     #[cfg(feature = "async")]
-    {
-      self.pool.inner.waiting_futures.wake_next();
-    }
+    self.pool.notify();
   }
 }
 
 impl<T> Lease<T> {
-  fn from_arc_mutex(arc: &Arc<Wrapper<T>>, #[allow(unused)] pool: &Pool<T>) -> Option<Self> {
+  fn from_arc_mutex(arc: &Arc<Mutex<T>>, #[allow(unused)] pool: &Pool<T>) -> Option<Self> {
     arc.try_lock().map(|guard| {
       std::mem::forget(guard);
       Self {
